; Camp.asm ( ASSEMBLER SOURCE CODE MODULE )
;       Installs a TSR that allocates a desired amount of memory 
;       and then remains resident until uninstalled. So it "camps" on memory.
;       This module is called by a C-program via 
;       _setup(unsigned par) where par gives the number of paragraphs
;       of memory to allocate in total. The C-program can also call 
;       void uninstall(void), which frees memory and otherwise cleans up
;       after the camper. (This would normally be a second invocation of camp
;       later on.
;
;       All three routines _uninstall, _setup, and _tsr below are non-resident.
;       tsr_ is called from the C-program, and it leaves behind the resident
;       code named _uninst.  

;       The only practical way to invoke a tsr is via an interrupt, but which
;       interrupt to use? Many of them are used by the system and almost all of
;       them could be used by some other program that might be run while we
;       are resident. We settled on 01H, which is normally only used by
;       debuggers. 
;
;	When it is wakened by 01H, _uninst first invokes the original
;       handler and waits for its return. In theory this should make the
;       presence of camp invisible to other programs, even ones that are
;       using int 01H. Then it merely
;       retruns three addresses in registers:  It's own PSP segment address
;       in si (used as a handle to free memory), the stored address of the
;       01H handler that was present when camp was installed (so it can
;       be restored) in cx:dx, and the entry point of _unist itself in
;       es:bp. 

;	When it is called from main, uninstall first issues int 01 in order
;       retrieve the information _uninst reports. Then it makes sure that the
;       current  01H 
;       vector matches the address of _uninst in case some other handler might 
;       have been installed in the meantime. If everything is ok, it then
;       restores the old handler and ask DOS to free the memory the resident
;       routine has been camping on. 


_TEXT SEGMENT PUBLIC 'CODE'
ASSUME cs:_TEXT,ds:_TEXT,ss:_TEXT
PUBLIC _quiet,_setup,_tsr,_uninstall

org 100h
begin:            ; Program entry point
jmp near ptr crt  ; Jump around to beginning of C-run time


;
; BEGINNING  OF RESIDENT ROUTINE
;       handle int 01h in order to send data back to
;       non-resident program for uninstall purposes.
;

_resident:

; Resident data
PSP_SEG dw ?           ; segment address of our PSP
uninst_seg_hi db ?    ; Used to store the seg:offset address of the uninst
uninst_seg_lo db ?    ; label in the tsr, which is the entry point for our
uninst_off_hi db ?    ; int 01 handler. Since segmented addresses are not
uninst_off_lo db ?    ; unique we need this exactly the way it was stored
old01vseg dw 00H
old01voff dw 00H      ; address of old int 01H handler: for uninstall

_uninst:              ; entry point of 01H handler
	 push ds    
	 mov ax,cs      ; Load ds so we can access our own data
	 mov ds,ax
;
; INT 01H HANDLER
;
; This routine is invoked from camp (new incarnation) via int01
; Its job is to return the address of the original int 01H handler
; in cx:dx 
; so that it can be restored in the uninstall routine of camp.
; Also sends cs:_uninst in bp:es
; so that it can be compared with the current interrupt vector
; for int 01H. If they don't agree, then some other routine
; has stolen the interrupt, and it's not safe to uninstall.
; Finally, it sends segment of PSP in si

;
; CHAIN TO OLD 01H HANDLER
;
	pushf                     ; The first 3 pushes get us back here when
	push cs                   ; we rti from old handler
	push offset back_again    ; cs:offset back_again is return  
	push word ptr [_resident + 6]        ; push old01vseg            
	push word ptr [_resident + 8]        ; push old01voff           
	retf                      ; Going to old handler now. Bye!

back_again:             
;
;  Now we are going to send back information in the registers, but we
;  wouldn't want to do this if some other program had taken over
;  int 01H -- it wouldn't be expecting what we are going to send.
;  So check to make sure that the current 01H interrupt handler points
;  to uninst.  
	cli                       ; turn off interrupts for now
	mov ah,0000:0007          ; seg of 01H handler
	cmp ah,byte ptr [_resident + 2]      ; compare with our saved uninst address
	jne no_can_do             ; and abort the uninstall if they are
	mov al,0000:0006          ; not the same.
	cmp al, byte ptr [_resident + 3] 
	jne no_can_do
	mov ah,0000:0005          ; offset of 01H handler
	cmp ah, byte ptr [_resident + 4]
	jne no_can_do
	mov al,0000:0004
	cmp al, byte ptr [_resident + 5]
	jne no_can_do
	jmp cont                  ; ok to continue with uninstall

no_can_do:
	sti
	pop ds      ; returning to somebody who invoked 01H. Debugger?
	iret

cont:   sti                  ; throw away: we're returning to our own program
	mov ax,PSP_SEG       ; load registers with all our info
	mov si,ax
	mov ax, word ptr [_resident + 8] ;  old01voff
	mov dx,ax
	mov ax, word ptr [_resident + 6] ;  old01vseg
	mov cx,ax
	mov ax,cs
	mov bp,ax
	mov ax,offset _uninst
	mov es,ax
	pop ds
	iret

endtsr:            ; end of resident routine 
;                   
;                   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Non Resident Data
err_msg_1    db 'There is something wrong. Cannot install.0DH,0AH','$'
err_msg_2    db 'Cannot uninstall. No, or wrong 01H vector present.',0DH,0AH,'$'
motto db '                         CHICKEN HAVEN SOFTWARE',0DH,0AH,'$' 
install db 'Camp 1.0 By Terry McConnell (2017) No rights reserved.',0DH,0AH,'$'
parlen db 16     ; for a div
_quiet db 0; flag to signal that operation should be quiet
par_req dw ?

;
; TRANSIENT ("SETUP") PART OF PROGRAM
; This gets called from a C-program
;
_setup  proc    ; beginning of "transient" routine
	push bp ; calibrate stack frame
	mov bp,sp
;
;  UNLOAD PARAMETER FROM THE STACK
;
	mov ax,[bp+4]       ; get parameter from C program (16 bit int)
	mov par_req,ax       ; store par requested (1 word)
;
;  SAVE SEGMENT ADDRESS OF OUR PSP (for uninstall)
;
	mov ah,51H        ; Dos function: Get PSP address
	int 21H
	mov PSP_SEG,bx    ;segment address of current PSP
;
;  SAVE ADDRESS OF OLD INTERRUPT 01H HANDLER (for uninstall)
;

	mov al,01H  ; save address of old 01H handler
	mov ah,35H   ; using dos function 35h
	int 21H
	cli
	mov ax,es
	mov old01vseg,ax   ; save segment register of handler
	mov ax,bx
	mov old01voff,ax   ; save offset of handler
	sti
;
; INSTALL NEW HANDLER FOR 01H ( the resident part of this program: uninst)

        mov dx, offset _uninst
        mov al, 01H
        mov ah,25H
        int 21H

	cli
	mov al,0000:0007
	mov uninst_seg_hi,al
	mov al,0000:0006
	mov uninst_seg_lo,al
	mov al, 0000:0005
	mov uninst_off_hi,al
	mov al, 0000:0004
	mov uninst_off_lo,al
	sti

        cmp _quiet,0
        jne gohome
	mov dx, offset motto
	mov ah,09H
	int 21H           ;Print our motto
	mov dx, offset install
	mov ah,09H
	int 21H           ; Print our copyright

gohome: 
	pop bp            ; restore bp
	ret

barf:       
	mov dx,offset err_msg_1
	mov ah,09H    ; display string DOS service
	int 21H       ; DOS
	pop bp         ; restore bp
	mov al,01H     ; return value signals type 1 error
	mov ah,4CH    ; return to DOS without installing TSR.
	int 21H         ; exit to DOS

_setup   endp

; tsr:
; sets up a terminate and stay resident program to camp out

_tsr proc near

; Free the program's environment block. The tsr does not use it

        push es
	push bx
        mov ax, PSP_SEG
        mov es,ax        ; store psp seg address in es
        mov ax, es:002CH  ;seg address of env block at offset 2CH in psp
        mov es,ax
        mov ah,49H        ; free allocated memory (environment block)
        int 21H

	mov ax,PSP_SEG
	mov es,ax
	mov bx,par_req ; 
	mov ax,offset endtsr  ; compute number of paragraphs of resident
	div parlen     ; divide by 16 = paragraph length in bytes
        xor ah,ah      ; clear remainder
        add ax,12h     ; convert to offset from PSP
	cmp ax,bx      ; is it less than requested ?
	jb set_block_size ; ok, if so
        mov bx,ax      ; else, keep minimum required to hold resident part
set_block_size:
	mov ah, 4AH    ; es = PSP address, bx = pars in resize
	int 21H
	mov dx,bx
	pop bx
	pop es
        xor al,al      ; make the return value 0
        mov ah,31h     ; keep program, return to DOS, keep dx pars
	int 21H
_tsr endp

;
;  UNINSTALL ROUTINE
;

;  The installed tsr contains an int 01H handler whose purpose is to
;  return some stored data in the registers. The data returned are
;  si: segment address of the resident program's PSP
;  cx:dx address of the original int 01H handler
;  bp:es address of resident program's own int 01H handler
;  This routine refuses to uninstall the tsr if the
;  current handler address 
;  is different from our own interrupt vector
;  handler assignment-- that would mean some other handler had been
;  installed in the meantime. Assuming that one chains back to ours,
;  we had better not remove it. (Hint from Marc Johnson.)

_uninstall proc near
	push ds
	push si
	push es
	push bp
	push cx
	push dx

	mov al,01H  ; save address of current 01h handler as es:bx
	mov ah,35H   ; using dos function 35h
	int 21H
	mov ax,es
	mov old01vseg,ax   ; save segment register of handler
	mov ax,bx
	mov old01voff,ax   ; save offset of handler

	;  OK, now we can uninstall our handler


; We must put everything back the way we found it: the old 
; int01h vector must be restored and the program memory must be freed.

	int 01H   ; query resident routine for info
	mov ax,es  
	cmp ax, [old01voff]
        jne err2
	mov ax,bp
	cmp ax, [old01vseg]
        jne err2
	mov ax,cx       ; ds:dx wants address of handler to restore
	mov ds,ax
	mov al,01H
	mov ah,25H
	int 21H         ; restore original int 01H handler
	mov ax,si
	mov es,ax     ; seg address of psp
	mov ah,49H
	int 21H       ; free allocated memory
byebye:
	pop dx
	pop cx
	pop bp
	pop es
	pop si
	pop ds
	ret           ; to main ( where a success message is printed.)

err2:    mov dx, offset err_msg_2          ; print an error message and exit
	mov ah,09H
	int 21H
	mov al,01H    ; return value is error 1
	mov ah,4CH
	int 21H       ; return to DOS
_uninstall endp

crt:       ; Borland C run-time stuff goes below
_TEXT ends


end begin
